#include <Servo.h>
#include "DHT.h"
#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include <SoftwareSerial.h>
#include <TimeLib.h>


// note muzicale pentru crearea melodiei la intrarea in casa
#define NOTE_C4  262
#define NOTE_E4  330
#define NOTE_G4  392
#define NOTE_A3  220
#define NOTE_G3  196
#define NOTE_B3  247
#define NOTE_D4  294

// variabile utilizate pentru aprinderea ledului in functie de 
// sunet, luminozitate și numărul de secunde 
int relay_close_led_flag;
int relay_close_led_count;

// variabila care contorizeaza câți oameni intră în casă
int counter_people;

// variabilă care, atunci când devine multiplu de 3, activează 
// alarma casei
int counter_police;

// variabila care opreste alarma dupa un anumit timp
int cate_sec = 0;
int i;

// variabila care contorizeaza câte încercări mai are utilizatorul de ghicire
// a parolei 
int tries = 3;

// utilizați în comunicarea serială
char txbuf[20];
char rxbuf[20];
int index = 0;
long ts = 0;

// aprinde și stinge LED-ul după un interval
short blink_on = 0;
int ok;

// activare/dezactivare alarmă
int ok1 = 0;

// variabile pentru reținerea luminozității/zgomotului extern
volatile int Brightness;
volatile int Environmental_sound;

// variabilă care permite sau nu intrarea în garaj
int pass;

// servomotoarele
Servo servo_9;
Servo servo_10;
Servo servo_11;

// variabilă care reține nivelul apei pentru o plantă
int water_level;
// variabilă care reține umiditatea solului unei plante
int Soil_moisture;

// variabilă care detectează dacă este sau nu foc 
int flame;

// variabile care contorizează câte grade sunt afară, respectiv care este
// umiditatea mediului
int temperature;
int humidity;
DHT dht12(12, 11);

// butoane folosite în scrierea parolei
int Button_Red;
int Button_Green;

// contorizare de câte ori a fost apăsat un anumit buton
int Red_num_time;
int Green_num_time;

// parola casei
String password;

// variabile care deschid sau închid ușa, în funcție de parolă sau dacă
// este detectat sau nu incendiu
int door_flag;
boolean key_door_flag;

// LCD
LiquidCrystal_I2C mylcd(0x27,16,2);

// variabile care ajută în conectarea aplicației pentru ușă, mașină, LED, geam,
// parcare, parolă
int BLE_door_flag;
int BLE_car_flag;
String BLE_value;
int BLE_LED_flag;
int BLE_window_flag;
int BLE_fan_flag;
String sensor_data;
String app_password;
int re_string_len;


// funcție auxiliară pentru a aprinde și stinge LED-ul în cazul sistemului
// de udare al plantelor
void soil() {
  if(Serial.available()) {
    char a = Serial.read();
    if (a == '\n') {
      rxbuf[index] = '\0';
      // on = se pornește sistemul de udare al plantelor
      if (strcmp(rxbuf, "on") == 0) {
        // se poate aprinde LED-ul
        ok = 1;
        blink_on = !blink_on;
        delay (500);
        PORTB &= ~(1 << PB5);
        // alarma pornește un zgomot de înștiințare că planta nu are 
        // suficient
        tone(3,532);
      }
    }

    // se continuă scrierea și timiterea mesajelor pe Serial
    else {
      rxbuf[index] = a;
      index++;
      rxbuf[index] = '\0';
    }
  }

  // blink-uirea LED-ului
  if (blink_on) {
      if ((millis() - ts) >= 100) {
        ts = millis();
        PORTB ^= (1 << PB5);
      }
  }
}

// blink-uirea unui LED în caz de furtună
void Storm_warning()  {
  // aprindere LED
  digitalWrite(7,HIGH);
  delay(125);
  delay(200);

  // stingere LED după un delay de 325
  digitalWrite(7,LOW);
  delay(200);
}

// funcție care pornește ventilatorul în caz de incendiu/temperatură prea ridicată
void fan_motor_pwm(int speedpin, int speed) {

  // nu se învârte ventilatorul
  if (speed <= 0) {
    analogWrite(speedpin, 0);
  }

  // are viteză prea mare, o ajustăm
  else if (speed > 255) {
    analogWrite(speedpin, 255);
  }

  // se învârte cu viteza specificată
  else {
    analogWrite(speedpin,speed);
  }
}

// se definește funcția de sunet pentru butoanele ce sunt apăsate
// la introducerea parolei
void key_voice() {
  tone(3,349);
  delay(125);
  delay(100);
  noTone(3);
  delay(100);
}

// daca parola = 123 => corect. Altfel, parola este greșită, numărul de încercări de ghicire a parolei scade și
// contorul pentru alarmă crește
void Password_Confirmation() {

  // se apasă butonul verde => se încearcă completarea parolei, fără validare
  if (Button_Green == 0 && Button_Red != 0) {
    delay(100);

    // butonul roșu validează parola
    Button_Red = digitalRead(5); 
    if (Button_Green == 0 && Button_Red != 0) {
      // ușa este încă închisă
      if (key_door_flag == false) {
        // parola a fost introdusă corect
        if (password == "123") {

          // crește numărul oamenilor care intră în casă
          counter_people++;

          // se afișează pe ecran mesaje corespunzătoare acțiunilor
          mylcd.clear();
          mylcd.setCursor(1-1, 1-1);
          mylcd.print("   Welcome!");
          mylcd.setCursor(1-1, 2-1);
          mylcd.print("  People nr: ");
          mylcd.setCursor(14-1, 2-1);
          mylcd.print(counter_people);
          servo_9.write(180);
          delay(0);
          // după terminarea melodiei, se închide ușa
          door_flag = 0;

          // se deschide ușa
          key_door_flag = true;
          // se pornește melodia de intrare în casă
          setup_melody_door();

        } else {

          // parola nu a fost introdusă corect
          mylcd.clear();
          mylcd.setCursor(1-1, 1-1);
          mylcd.print("Attempts Left:");

          // numărul de încercări rămase scade
          tries--;
          mylcd.setCursor(8-1, 2-1);
          // se afișează pe ecran câte încercări mai are utilizatorul până la 
          // pornirea alarmei
          mylcd.print(tries);

          // de fiecare dată când numărul de încercări rămase scade, se contorizează 
          // cât de aproape este de a chema poliția
          counter_police ++;
          if(counter_police % 3 == 0) {

            // poliția este chemată la fiecare 3 încercări greșite, iar dup ce aceasta este 
            // chemată și alarma este pusă în funcțiune, numprul de încercări devine din nou 3
            counter_police = 0;
            ok1 = 1;
          }
          
          // poliția este chemată
          while(ok1 == 1)
           {
             // pornește alarma casei care cheamă poliția
             for(i = 700; i < 800; i++) {   
                  tone(3,i);
                  delay(15);
              } 
              for(i = 800; i > 700; i--){
                tone(3,i);
                delay(15);
              }
              delay(1000);

              // contorizare câte secunde au trecut de la chemarea poliției
                cate_sec ++;

                // afișare mesaj corespunzător pe ecran
                mylcd.clear();
                mylcd.setCursor(1-1, 1-1);
                mylcd.print("   BLOCKED");
                mylcd.setCursor(1-1, 2-1);
                mylcd.print(" Calling Police");

                // dacă au trrecut 3 secunde de la pornirea alarmei, aceasta se poate opri
                // întrucât poliția a fost deja chemată
                if (cate_sec == 3) { 
                    ok1 = 0;            
                    tries = 3;
                 }
           }
        // mesaje corespunzătare pentru o parolă greșită
          tone(3,165);
          delay(125);
          delay(800);
          noTone(3);
          delay(200);
          mylcd.clear();
          mylcd.setCursor(1-1, 1-1);
          mylcd.print("   Taisia's Home");
          mylcd.setCursor(11-1, 2-1);
          mylcd.print("Again ");
          mylcd.setCursor(1-1, 2-1);
          mylcd.print("Password:");
      }

      } else if (true == key_door_flag) {

        // când se intră în casă, se afișează un mesaj corespunzător și temperatura
        key_door_flag = false;
        mylcd.clear();
        mylcd.setCursor(1-1, 1-1);
        mylcd.setCursor(1-1, 1-1);
        mylcd.print("Temperature:");
        mylcd.setCursor(13-1, 1-1);
        mylcd.print(dht12.readTemperature());
        mylcd.setCursor(1-1, 2-1);
        mylcd.print("   I am home!");
        servo_9.write(90);
        delay(0);
      }
      password = "";

    }
  }
}
// definirea melodiei de intrare în casă
int tempo = 160;
int buzzer = 3;
int melody[] = {

    NOTE_C4, 4, NOTE_C4, 4, NOTE_E4, 8, NOTE_G4, 8, 
    NOTE_E4, 4, NOTE_G4, 4, NOTE_E4, 4, NOTE_E4, 4,
    NOTE_C4, 4, NOTE_C4, 8, NOTE_A3, 4, NOTE_C4, 8,
    NOTE_G3, 4, NOTE_B3, 4, NOTE_D4, 4, NOTE_B3, 4,
    NOTE_G3, -1    
};
int notes = sizeof(melody) / sizeof(melody[0]) / 2;
int wholenote = (60000 * 4) / tempo;
int divider = 0, noteDuration = 0;


void setup_melody_door() {
  for (int thisNote = 0; thisNote < notes * 2; thisNote = thisNote + 2) {
    divider = melody[thisNote + 1];
    if (divider > 0) {
      noteDuration = (wholenote) / divider;
    } else if (divider < 0) {
      noteDuration = (wholenote) / abs(divider);
      noteDuration *= 1.5;
    }
    tone(buzzer, melody[thisNote], noteDuration * 0.75);
    delay(noteDuration);
    noTone(buzzer);
  }
}

// funcție utilizată pentru afișarea corectă a datei și a orei la intrarea 
// în casă
void printDigits(int digits) {
  if (digits < 10) {
    mylcd.print("0");
  }
  mylcd.print(digits);
}

void setup() {
  BLE_value = "Empty";
  BLE_door_flag = 0;
  BLE_car_flag = 0;
  BLE_LED_flag = 0;
  BLE_window_flag = 0;
  BLE_fan_flag = 0;
  app_password = ' ';
  sensor_data = ' ';
  re_string_len = 0;

  counter_people = 0;
  counter_police = 0;

  relay_close_led_flag = 0;
  relay_close_led_count = 0;
  ok = 0;
  Brightness = 0;
  Environmental_sound = 0;
  Serial.begin(9600);
  delay(100);
  // initializare digital pin A1,D2 ca input.
  pinMode(A1, INPUT);
  pinMode(2, INPUT);
  // initializare digital pin D7 ca output.
  pinMode(7, OUTPUT);


  pass = 0;
  pinMode(A0, INPUT);
  // Definire poziție servomotor pe D11
  servo_11.attach(11);


  water_level = 0;
  pinMode(A2, INPUT);
  // Definire poziție servomotor pe D10
  servo_10.attach(10);


  Soil_moisture = 0;
  pinMode(13, OUTPUT);
  pinMode(3, OUTPUT);
  pinMode(A3, INPUT);

  flame = 0;
  pinMode(8, INPUT);
  pinMode(6, OUTPUT);
  digitalWrite(6, LOW);


  temperature = 0;
  humidity = 0;

  // inițializare senzor
  dht12.begin();
  pinMode(6, OUTPUT);
  digitalWrite(6, LOW);


  pinMode(3, OUTPUT);
  Button_Red = 1;
  Button_Green = 1;
  Red_num_time = 0;
  Green_num_time = 0;

  // parola este inițial nulă
  password = "";
  door_flag = 0;
  key_door_flag = false;
  mylcd.init();
  // afișare temperatură pe LCD
  mylcd.backlight();
  servo_9.attach(9);
  mylcd.setCursor(1-1, 1-1);
  mylcd.print("Temperature:");
  mylcd.setCursor(13-1, 1-1);
  mylcd.print(dht12.readTemperature());

  // data și ora corespunzătoare
  setTime(10, 24, 0, 26, 5, 2023);
  
  // obține informații despre data și ora curente
  time_t t = now();
  tmElements_t tm;
  breakTime(t, tm);

  // afișare dată și oră pe LCD
  mylcd.setCursor(0, 1);
  mylcd.print(tmYearToCalendar(tm.Year));
  mylcd.print("/");
  printDigits(tm.Month);
  mylcd.print("/");
  printDigits(tm.Day);
  mylcd.print(" ");
  printDigits(tm.Hour);
  mylcd.print(":");
  printDigits(tm.Minute);
  servo_9.write(90);
  delay(0);
  pinMode(5, INPUT);
  pinMode(4, INPUT);
}


void loop() {
   //Citire valoarea tensiunii analogice de ieșire a senzorului fotosensibil și salavre în variabila corespunzătoare
    Brightness = analogRead(A1);
    
    //Citirea stării de ieșire digitală a senzorului de sunet și salvare în variabila corespunzătoare
    Environmental_sound = digitalRead(2);
    // afișare pe serial cât este luminozitatea în afara casei
    // Serial.println(Brightness);

    // este întuneric afară
    if (Brightness > 550) {
      // și este gălăgie împrejurul casei
      if (Environmental_sound == 1) {
      // se aprinde LED-ul
      digitalWrite(7,HIGH);
      relay_close_led_flag = 1;

    } else {  // afară este lumină
      if (relay_close_led_flag == 1) {
        relay_close_led_count = relay_close_led_count + 1;
        // la 5 secunde, LED-ul blink-uiește
        if (relay_close_led_count >= 5000) {
          // LED-ul stă stins
          digitalWrite(7,LOW);
          relay_close_led_count = 0;
          relay_close_led_flag = 0;
        }
      }
    }
    } 
    else 
    {
      // stingerea LED-ului
        if (BLE_LED_flag != 1) {
          digitalWrite(7,LOW);
      }
    }
    // citire stare de introducere a monedei pentru parcare
    pass = digitalRead(A0);

    // s-a introdus o monedă
    if (pass == 0) {

      // servomotorul se poziționează așa încât mașina să poată intra în casă
      servo_11.write(180);

      // afișare mesaje corespunzătoare pe LCD
      mylcd.clear();
      mylcd.setCursor(1-1, 1-1);
      mylcd.setCursor(1-1, 1-1);
      mylcd.print(" Car is parking! ");
    } 

    // nu este permisă intrarea mașinii în garaj/ s-a finalizat procesul de
    // introducere a monedei
    else if (BLE_car_flag != 1 && pass == 1) 
    {
      // poziția de închidere a barierei
      servo_11.write(70);
    }

    // nivelul de apă detectat
    water_level = analogRead(A2);

    // dacă este detectat un nivel ridicat de apă, atunci fereastra se va închide
    if (water_level > 100) 
    {
      // închiderea ferestrei
      servo_10.write(90);
      // pornirea alarmei de furtună
      Storm_warning();
    } 
    else if (BLE_window_flag != 1)
    {
      // se deschide fereastra atunci când nu este niciun pericol
      servo_10.write(0);
    }
   
   // se pornește aplicația de monitorizare a solului
   soil();
   Soil_moisture = analogRead(A3);

    // afișare detalii pe LCD despre o anumită plantă ce este monitorizată
    if (! (50 > Soil_moisture && water_level < 100) && ok == 1) {
      digitalWrite(13,LOW);
      noTone(3);
      mylcd.clear();
      mylcd.setCursor(1-1, 1-1);
      mylcd.print("Moisture:");
      mylcd.setCursor(10-1, 1-1);
      mylcd.print(Soil_moisture);
      mylcd.setCursor(1-1, 2-1);
      mylcd.print("Water_lvl:");
      mylcd.setCursor(11-1, 2-1);
      mylcd.print(water_level);
    }

    // detectarea unei flăcări
    flame = digitalRead(8);
    if (1 == flame) {  // este detectat foc
      // se va aprinde ventilatorul
      fan_motor_pwm(6, 120);
      // se va deschide ușa pentru evacuare
      servo_9.write(180);
      // lumea este evacuată
      counter_people = 0;

      // se afișează un mesaj corespunzător pe LCD
      mylcd.clear();
      mylcd.setCursor(1-1, 1-1);
      mylcd.print("     FIRE!!");
      mylcd.setCursor(1-1, 2-1);

      // se cheamă ambulanța
      mylcd.print("  Calling Ambulance");
      // alarma corespunzătoare ambulanței
      tone(3,650,750);
      delay (300);
      tone(3,900,1000);
      delay (300);
      tone(3,650,750);
      delay (300);
      tone(3,900,1000);

    } 
    else {  // nu este detectat foc
      // nu se învârte ventilatorul și ușa este închisă
      if (BLE_fan_flag != 1) {
        fan_motor_pwm(6, 0);
        servo_9.write(90);
    } 
    }

    // determinarea umidității și a temperaturii din afara casei
  temperature = dht12.readTemperature();
  humidity = dht12.readHumidity();
  flame = digitalRead(8);
  // la o creștere a temperaturii, se va porni automat ventilatoruș
  if (temperature > 25) {
    fan_motor_pwm(6, 120);

  } else {
    // temperatura este una corespunzătoare => ventilatorul este închis 
    if (BLE_fan_flag != 1) {
      fan_motor_pwm(6, 0);
    }
  }
    // buton pentru scrierea parolei
    Button_Green = digitalRead(4);
    // buton pentru validarea parolei
    Button_Red = digitalRead(5);
    // este validată parola
    if (Button_Green != 0 && Button_Red == 0) {
      delay(100);
      Button_Green = digitalRead(4);
      while (Button_Green != 0 && Button_Red == 0) {
        Button_Red = digitalRead(5);
        // contorizare de câte ori s-a apăsat butonul roșu
        Red_num_time = Red_num_time + 1;
        delay(100);
      }
    }

    if (Button_Green == 0 && Button_Red != 0) {
      delay(100);
      Button_Red = digitalRead(5);
      while (Button_Green != 0 && Button_Red == 0) {
        Button_Green = digitalRead(5);
        // contorizare de câte ori s-a apăsat butonul verde
        Green_num_time = Green_num_time + 1;
        delay(100);
      }
    }

    // când se apasă mai mult de 15 u => se șterge un caracter 
    if(Red_num_time > 15) {
      key_voice();
      password = password.substring(0, password.length() - 1);
      mylcd.clear();
      mylcd.setCursor(1-1, 1-1);
      mylcd.print("   Taisia's Home");
      mylcd.setCursor(1-1, 2-1);
      mylcd.print("Password:");
      mylcd.setCursor(11-1, 2-1);
      mylcd.print(password);
    }

    // când se apasă între 1 și 4 u => echivalent cu afișarea caracterului 1
    if (1 < Red_num_time && 4 > Red_num_time) {
      key_voice();
      password = String(password) + String("1");
      mylcd.clear();
      mylcd.setCursor(1-1, 1-1);
      mylcd.print("   Taisia's Home");
      mylcd.setCursor(1-1, 2-1);
      mylcd.print("Password:");
      mylcd.setCursor(11-1, 2-1);
      mylcd.print(password);
    }

    // când se apasă între 4 și 9 u => echivalent cu afișarea caracterului 2
    if (4 < Red_num_time && 9 > Red_num_time) {
      key_voice();
      password = String(password) + String("2");
      mylcd.clear();
      mylcd.setCursor(1-1, 1-1);
      mylcd.print("   Taisia's Home");
      mylcd.setCursor(1-1, 2-1);
      mylcd.print("Password:");
      mylcd.setCursor(11-1, 2-1);
      mylcd.print(password);
    }

    // când se apasă între 9 și 15 u => echivalent cu afișarea caracterului 3
    if (9 < Red_num_time && Red_num_time < 15) {
      key_voice();
      password = String(password) + String("3");
      mylcd.clear();
      mylcd.setCursor(1-1, 1-1);
      mylcd.print("   Taisia's Home");
      mylcd.setCursor(1-1, 2-1);
      mylcd.print("Password:");
      mylcd.setCursor(11-1, 2-1);
      mylcd.print(password);
      
    }

    // apelare funcție de confirmare a parolei 
    Password_Confirmation();
    Red_num_time = 0;

      // cod generat de aplicația Mixly pentru a putea folosi aplicația
      BLE_value = Serial.readString();
      if (12 > String(BLE_value).length()) {
        if ('%' == String(BLE_value).charAt(0) && '#' == String(BLE_value).charAt(2)) {
          switch (String(BLE_value).charAt(1)) {

          // validează dacă parola este corectă sau nu
           case 'Q':
            re_string_len = String(BLE_value).length();
            app_password = String(BLE_value).substring(3,(re_string_len - 2));
            if (app_password == "123") {
              servo_9.write(180);
              delay(0);
            
              Serial.println("%A#");

            } else {
              Serial.println("%B#");
            }
            break;
            // afișează informațiile referitoare la mediul ambiental
           case 'K':
            sensor_data = String("%") + String(String(Brightness) + String(String("%") + String(String(String(temperature)) + String(String("%") + String(String(String(humidity)) + String(String("%") + String(String(String(Soil_moisture)) + String(String("%") + String(String(String(water_level)) + String("%"))))))))));
            Serial.println(sensor_data);
            break;
            // aprindere și stingere LED
           case 'A':
            BLE_LED_flag = 1;
            digitalWrite(7,HIGH);
            break;
           case 'B':
            BLE_LED_flag = 0;
            digitalWrite(7,LOW);
            break;
            // deschidere și închidere ușă
           case 'C':
            BLE_door_flag = 1;
            servo_9.write(180);
            delay(0);
            break;
           case 'D':
            BLE_door_flag = 0;
            servo_9.write(90);
            delay(0);
            break;
            // aprindere și închidere ventilator
           case 'E':
            BLE_fan_flag = 1;
            fan_motor_pwm(6, 120);
            break;
           case 'F':
            BLE_fan_flag = 0;
            fan_motor_pwm(6, 0);
            break;

            // deschidere și închidere garaj
           case 'I':
            BLE_car_flag = 1;
            servo_11.write(180);
            delay(0);
            break;
           case 'J':
            BLE_car_flag = 0;
            servo_11.write(90);
            delay(0);
            break;

            // deschidere și închidere geam
           case 'L':
            BLE_window_flag = 0;
            servo_10.write(0);
            delay(0);
            break;
           case 'M':
            BLE_window_flag = 1;
            servo_10.write(90);
            delay(0);
            break;
          }
          BLE_value = "Empty";

        } else {
          BLE_value = "Empty";

        }

    }
}
